%{
/*
libfawk - A function-only AWK dialect (BASIC compiler)

Copyright (c) 2019 Tibor 'Igor2' Palinkas. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  * 1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
  * 3. Neither the name of the copyright holder nor the names of its
    contributors may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Project page: http://repo.hu/projects/libfawk
Source code: svn://repo.hu/libfawk/trunk
Contact the author: http://igor2.repo.hu/contact.html
*/

#define fbas_ctx_t fawk_ctx_t

#include "config.h"
#include <stdlib.h>
#include <stdio.h>
#include <genht/hash.h>
#include "libfawk.h"
#include "parser.h"
#include "fbas.y.tab.h"
#include "parser_fbas.h"
#include "symtab.h"
#include "compile.h"
#include "execute.h"

#define PUSH_IP() fawk_push_num(ctx, ctx->code.used)
#define CURR_IP() ctx->code.used

#include "label_inlines.h"
#include "mainprog_inlines.h"

%}

%union {
	char *str;
	fawk_num_t num;
}

%token BT_NEQ BT_GTEQ BT_LTEQ
%token BT_NIL
%token BT_IF BT_THEN BT_ELSE BT_FOR BT_TO BT_STEP BT_NEXT
%token BT_DO BT_WHILE BT_UNTIL BT_LOOP
%token BT_GOTO BT_DEF BT_END BT_LET BT_CURRFUNC

%token <str> BT_ID BT_STRING
%token <num> BT_NUM

%left BT_OR
%left BT_AND
%left BT_IN
%right '='
%left BT_NEQ BT_LTEQ '<' BT_GTEQ '>'
%left '@'
%left '+' '-'
%left '*' '/' BT_MOD
%left BT_NOT
%right '(' ')'
%%

script:
	{
		if (bas_init_labels(ctx) != 0) return -1;

		/* the first thing in a BASIC program is main() - emulate it */
		fawk_symtab_regfunc(ctx, "main", CURR_IP(), 1, -1);
		
		ctx->fp = ctx->sp;
	}
	maybe_newlines
	program
	{
		bas_end_main(ctx);
		if (bas_uninit_labels(ctx) != 0) return -1;
	}
	;

program:
	  func_def newlines program
	| line program
	| BT_END newlines { bas_end_main(ctx); } program
	| /* empty */
	;

line:
	  lineno_label statement sep
	| statement sep
	;

sep:
	':' | newlines
	;

newlines:
	  '\n'
	| newlines '\n'
	;

maybe_newlines:
	newlines | /* empty */;

lineno_label:
	  BT_NUM      { bas_add_label(ctx, NULL, $1); }
	| BT_ID ':'   { bas_add_label(ctx, $1, 0); }
	;

statement:
	  s_goto
	| e_call      { fawkc_addi(ctx, FAWKI_POP); /* void experssion's result value shall be removed from the stack */ }
	| s_assign
	| s_if
	| s_for
	| s_loops
	| s_return
	;

s_goto:
	  BT_GOTO BT_ID  { bas_add_jump(ctx, $2, 0); }
	| BT_GOTO BT_NUM { bas_add_jump(ctx, NULL, $2); }
	;

s_assign:
	  BT_LET assign
	|        assign
;

assign:
	var '=' expr { fawkc_addi(ctx, FAWKI_SET); fawkc_addi(ctx, FAWKI_PUSH_SYMVAL); fawkc_addi(ctx, FAWKI_POP); /* void expr*/ }
	;

/* The top of the stack between two statements is the function return value; pop it and allow expr to replace it */
s_return:
	  BT_LET BT_CURRFUNC '=' { fawkc_addi(ctx, FAWKI_POP); } expr
	|        BT_CURRFUNC '=' { fawkc_addi(ctx, FAWKI_POP); } expr
	;

s_if:
		if_then statement
			{
				size_t jmp1 = fawk_pop_num(ctx, 1);
				ctx->code.code[jmp1].data.num = CURR_IP();
			}

		| if_then
		newlines func_body
			{
				size_t jmp1 = fawk_pop_num(ctx, 1);
				ctx->code.code[jmp1].data.num = CURR_IP();
			}
		BT_END BT_IF
		| if_then
		newlines func_body
		BT_ELSE
		{
			fawkc_addi(ctx, FAWKI_JMP);
			PUSH_IP(); fawkc_addnum(ctx, 777);
		}
		newlines func_body
		{
			size_t jmp_then_post = fawk_pop_num(ctx, 1), jmp_if = fawk_pop_num(ctx, 1);
			ctx->code.code[jmp_then_post].data.num = CURR_IP();
			ctx->code.code[jmp_if].data.num = jmp_then_post+1;
		}
		BT_END BT_IF
	;

if_then:
	BT_IF expr BT_THEN    { fawkc_addi(ctx, FAWKI_POPJZ); PUSH_IP(); fawkc_addnum(ctx, 777); }
	;

s_for:
	s_for_in
	| s_for_plain
	;

s_for_in:
	BT_FOR var BT_IN var newlines
			{
				fawkc_addi(ctx, FAWKI_FORIN_FIRST);
				fawkc_addi(ctx, FAWKI_JMP); /* skip statements first, so next can load the variable first */
				fawkc_addnum(ctx, 0);
				PUSH_IP();
			}
		func_body
			{
				size_t begin = fawk_pop_num(ctx, 1);
				fawkc_addi(ctx, FAWKI_FORIN_NEXT);
				fawkc_addnum(ctx, begin);
				ctx->code.code[begin-1].data.num = CURR_IP()-2;
			}
		for_next
	;

s_for_plain:
	BT_FOR BT_ID '='
			{
				fawkc_addi(ctx, FAWKI_MAKE_SYMREF);
				fawkc_addsymref(ctx, $2, 0, ctx->compiler.funcdef_offs);
				fawkc_addnum(ctx, 0);
			}
			expr
			{
				fawkc_addi(ctx, FAWKI_SET);
				fawkc_addi(ctx, FAWKI_PUSH_SYMVAL);
				fawkc_addi(ctx, FAWKI_POP); /* void expr */

				fawkc_addi(ctx, FAWKI_JMP); /* jump-in: in the first iteration, skip condition and step and start at the body  */
				PUSH_IP(); /* slot A */
				fawkc_addnum(ctx, 0);

				PUSH_IP(); /* slot B */
			}
			BT_TO expr
			{
				/* check range : if ((to-VAR)*step) >= 0, we can continue  - it's really  ((to-(VAR-step))*step) >= 0 because we step only after deciding */

				/* 'to' expr is already on the stack get the VAR and sub from 'to' */
				fawkc_addi(ctx, FAWKI_MAKE_SYMREF);
				fawkc_addsymref(ctx, $2, 0, ctx->compiler.funcdef_offs);
				fawkc_addnum(ctx, 0);
				fawkc_addi(ctx, FAWKI_PUSH_SYMVAL);
				fawkc_addi(ctx, FAWKI_SUB);
			}
			for_step    /* 3rd expr - shall be run at the end */
			{
				/* need to dup the top 2 cells for the mul so step can be preserved for the actual stepping, plus swap them for adjusting VAR with step */
				fawkc_addi(ctx, FAWKI_PUSH_REL); fawkc_addnum(ctx, -1);
				fawkc_addi(ctx, FAWKI_PUSH_REL); fawkc_addnum(ctx, -3);

				/* step and (to-VAR) are on the stack */
				fawkc_addi(ctx, FAWKI_PUSH_REL); fawkc_addnum(ctx, -2);
				fawkc_addi(ctx, FAWKI_SUB);

				/* step and (to-(VAR-step)) are on the stack */
				fawkc_addi(ctx, FAWKI_MUL);

				/* push 0; stack is ((to-(VAR-step))*step),0 now */
				fawkc_addi(ctx, FAWKI_PUSH_NUM); fawkc_addnum(ctx, 0);

				fawkc_addi(ctx, FAWKI_GTEQ);
				fawkc_addi(ctx, FAWKI_POPJZ); /* jump out on false */
				PUSH_IP(); /* slot C */

				fawkc_addnum(ctx, 0);

				/* step; stack is [1-] = step, [-2] = ((to-(VAR-step))*step) */

				fawkc_addi(ctx, FAWKI_MAKE_SYMREF);
				fawkc_addsymref(ctx, $2, 0, ctx->compiler.funcdef_offs);
				fawkc_addnum(ctx, 0);
				fawkc_addi(ctx, FAWKI_PUSH_TOPVAR);
				fawkc_addi(ctx, FAWKI_PUSH_REL); fawkc_addnum(ctx, -3);  /* [-1]=VAR, [-2]=step, [-3]=... */
				fawkc_addi(ctx, FAWKI_ADD); fawkc_addi(ctx, FAWKI_SET);

				fawkc_addi(ctx, FAWKI_POP); /* void expr (stepping) */
				fawkc_addi(ctx, FAWKI_POP); /* remove step */
				fawkc_addi(ctx, FAWKI_POP); /* remove the original ((to-VAR)*step)*/

				fawkc_addi(ctx, FAWKI_JMP); /* skip 3rd expr */
				PUSH_IP(); /* slot D */
				fawkc_addnum(ctx, 0);


				fawkc_addi(ctx, FAWKI_JMP); /* jump back to the check */
				PUSH_IP(); /* slot E */
				fawkc_addnum(ctx, 0);
			}
			newlines func_body for_next
			{
				size_t jumpin, jback3rd, jskip, jout, rerun;

				jback3rd = fawk_pop_num(ctx, 1); /* slot E */
				jskip = fawk_pop_num(ctx, 1);    /* slot D */
				jout = fawk_pop_num(ctx, 1);     /* slot C */
				rerun = fawk_pop_num(ctx, 1);    /* slot B */
				jumpin = fawk_pop_num(ctx, 1);   /* slot A */

				fawkc_addi(ctx, FAWKI_JMP); /* jump back to execute the 3rd expr */
				fawkc_addnum(ctx, jskip+1);

				ctx->code.code[jumpin].data.num = jback3rd+1;
				ctx->code.code[jback3rd].data.num = rerun;
				ctx->code.code[jskip].data.num = jback3rd+1;
				ctx->code.code[jout].data.num = CURR_IP();
				fawkc_addi(ctx, FAWKI_POP);
				fawkc_addi(ctx, FAWKI_POP);
			}
	;

for_step:
	  /* empty */ { fawkc_addi(ctx, FAWKI_PUSH_NUM); fawkc_addnum(ctx, 1); }
	| BT_STEP expr
	;

for_next:
	  BT_NEXT
	| BT_NEXT var     { fawkc_addi(ctx, FAWKI_POP); /* void experssion */ }
	;

s_loops:
	BT_DO { PUSH_IP(); /*jump back here to repeat */ } s_loop_body;

s_loop_body:
	  s_while_loop
	| s_until_loop
	| s_loop_while
	| s_loop_until
	;

s_while_loop:
	BT_WHILE expr newlines           { fawkc_addi(ctx, FAWKI_POPJZ); PUSH_IP(); fawkc_addnum(ctx, 0); }
		func_body
		BT_LOOP                        { loop_pretest_jumpback(); }
		;

s_until_loop:
	BT_UNTIL expr newlines           { fawkc_addi(ctx, FAWKI_POPJNZ); PUSH_IP(); fawkc_addnum(ctx, 0); }
		func_body
		BT_LOOP                        { loop_pretest_jumpback(); }
		;

s_loop_while:
	newlines func_body
		BT_LOOP BT_WHILE expr          { fawkc_addi(ctx, FAWKI_POPJNZ); fawkc_addnum(ctx, fawk_pop_num(ctx, 1));}
		;

s_loop_until:
	newlines func_body
		BT_LOOP BT_UNTIL expr          { fawkc_addi(ctx, FAWKI_POPJZ); fawkc_addnum(ctx, fawk_pop_num(ctx, 1));}
		;


expr:
	  BT_NUM                      { fawkc_addi(ctx, FAWKI_PUSH_NUM); fawkc_addnum(ctx, $1); }
	| BT_STRING                   { fawkc_addi(ctx, FAWKI_PUSH_STR); fawkc_adds(ctx, $1); fawk_free(ctx, $1); }
	| BT_NIL                      { fawkc_addi(ctx, FAWKI_PUSH_NIL); }
	| var                         { fawkc_addi(ctx, FAWKI_PUSH_SYMVAL); }
	| '&' var                     /* done in var */
	| e_call
	| '(' expr ')'                /* done implicitly by the parser */
	| expr '=' expr               { fawkc_addi(ctx, FAWKI_EQ); }
	| expr BT_NEQ expr            { fawkc_addi(ctx, FAWKI_NEQ); }
	| expr BT_GTEQ expr           { fawkc_addi(ctx, FAWKI_GTEQ); }
	| expr BT_LTEQ expr           { fawkc_addi(ctx, FAWKI_LTEQ); }
	| expr '>' expr               { fawkc_addi(ctx, FAWKI_GT); }
	| expr '<' expr               { fawkc_addi(ctx, FAWKI_LT); }
	| expr BT_IN var              { fawkc_addi(ctx, FAWKI_IN); }
	| e_or
	| e_and
	| expr '+' expr               { fawkc_addi(ctx, FAWKI_ADD); }
	| expr '-' expr               { fawkc_addi(ctx, FAWKI_SUB); }
	| expr '*' expr               { fawkc_addi(ctx, FAWKI_MUL); }
	| expr '/' expr               { fawkc_addi(ctx, FAWKI_DIV); }
	| expr BT_MOD expr            { fawkc_addi(ctx, FAWKI_MOD); }
	| '-' expr                    { fawkc_addi(ctx, FAWKI_NEG); }
	| '+' expr                    /* a positive number is the same number, don't do anything */
	| BT_NOT expr                 { fawkc_addi(ctx, FAWKI_NOT); }
	| expr '@' expr               { fawkc_addi(ctx, FAWKI_CONCAT); }
	;

var:
	BT_ID          { fawk_push_num(ctx, ctx->compiler.numidx); ctx->compiler.numidx = 0; }
		arridx
		{
			fawkc_addi(ctx, FAWKI_MAKE_SYMREF);
			fawkc_addsymref(ctx, $1, ctx->compiler.numidx, ctx->compiler.funcdef_offs);
			fawkc_addnum(ctx, ctx->compiler.numidx);
			fawk_free(ctx, $1);
			ctx->compiler.numidx = fawk_pop_num(ctx, 1);
		}
	;

arridx:
	  /* empty */
	| arridx '[' ']'           { ctx->compiler.numidx = -1; }
	| arridx '[' aix_expr ']'  { ctx->compiler.numidx++; }
	| arridx '.' BT_ID         { ctx->compiler.numidx++; fawkc_addi(ctx, FAWKI_PUSH_STR); fawkc_adds(ctx, $3); fawk_free(ctx, $3); }
	| arridx '.' BT_STRING     { ctx->compiler.numidx++; fawkc_addi(ctx, FAWKI_PUSH_STR); fawkc_adds(ctx, $3); fawk_free(ctx, $3); }
	;

aix_expr: /* array index expression */
	  expr
	| aix_expr ','  { parse_aix_expr(); }
		expr          { fawkc_addi(ctx, FAWKI_CONCAT); }
	;

/* lazy OR evaluation */
e_or:
	expr BT_OR             { lazy_binop1(ctx, 1); }
		expr                 { lazy_binop2(ctx, 1); }
	;

/* lazy AND evaluation */
e_and:
	expr BT_AND             { lazy_binop1(ctx, 0); }
		expr                  { lazy_binop2(ctx, 0); }
	;

e_call:
	var '('                      { fawk_push_num(ctx, ctx->compiler.numargs); ctx->compiler.numargs = 0; }
		call_arg ')'
		{
			size_t old_numargs;

			old_numargs = fawk_pop_num(ctx, 1);

			fawkc_addi(ctx, FAWKI_CALL);
			fawkc_addnum(ctx, ctx->compiler.numargs);
			ctx->compiler.numargs = old_numargs;
		}
	;

call_arg:
	  /* empty */
	| expr                 { ctx->compiler.numargs++; }
	| expr ',' call_arg    { ctx->compiler.numargs++; }
	;

func_def:
	{
		fawk_cell_t *lc = fawk_push_alloc(ctx), *lc2 = fawk_push_alloc(ctx);
		lc->data.arr = (fawk_arr_t *)ctx->compiler.labels;
		lc2->data.arr = (fawk_arr_t *)ctx->compiler.lablink;
		if (bas_init_labels(ctx) != 0) return -1;
		fawk_push_num(ctx, ctx->fp);  /* we are really in main, that has an fp and numargs; save and reset it */
		fawkc_addi(ctx, FAWKI_JMP); PUSH_IP(); fawkc_addnum(ctx, 777); /* we are really in main() - need to jump over the function */
		ctx->fp = ctx->sp;
		ctx->compiler.funcdef_offs = ctx->sp;
	}
	BT_DEF BT_ID                         { ctx->compiler.numfixedargs = -1; ctx->compiler.numargs = 0; }
		'(' arg_def ')'                    { fawk_symtab_regfunc(ctx, $3, CURR_IP(), ctx->compiler.numargs, ctx->compiler.numfixedargs); ctx->fp = ctx->sp; ctx->parser.curr_func = $3; fawkc_addi(ctx, FAWKI_PUSH_NIL); /* this is going to be the return value */ }
		newlines
	func_body
	BT_END BT_DEF
	{
		int res;
		size_t jmp_over, n;
		fawk_cell_t cell;

		/* epilogue */
		fawkc_addi(ctx, FAWKI_RET);
		fawkc_addnum(ctx, ctx->compiler.numargs + (ctx->compiler.numfixedargs >= 0));
		for(n = 0; n < ctx->compiler.numargs + (ctx->compiler.numfixedargs >= 0); n++) {
			fawk_cell_t cell;
			fawk_pop(ctx, &cell);
			fawk_cell_free(ctx, &cell);
		}

		ctx->fp = 0;
		jmp_over = fawk_pop_num(ctx, 1); /* go back to set the jump-over address */
		ctx->code.code[jmp_over].data.num = CURR_IP();
		ctx->fp = fawk_pop_num(ctx, 1);
		ctx->compiler.funcdef_offs = 0;

		res = bas_uninit_labels(ctx);
		fawk_pop(ctx, &cell); ctx->compiler.lablink = (htpp_t *)cell.data.arr;
		fawk_pop(ctx, &cell); ctx->compiler.labels = (htpp_t *)cell.data.arr;
		free(ctx->parser.curr_func); ctx->parser.curr_func = NULL;
		if (res != 0) return -1;
	}
	;

arg_def:
	  /* empty */
	| BT_ID                                      { fawk_push_str(ctx, $1); fawk_free(ctx, $1); ctx->compiler.numargs++; }
	| '.' '.' '.'                                { ctx->compiler.numfixedargs = ctx->compiler.numargs; fawk_push_str(ctx, "VARARG"); }
	| arg_def ',' BT_ID                          { fawk_push_str(ctx, $3); fawk_free(ctx, $3); ctx->compiler.numargs++; }
	| arg_def ',' '.' '.' '.'                    { ctx->compiler.numfixedargs = ctx->compiler.numargs; fawk_push_str(ctx, "VARARG"); }
	;

func_body:
	  line func_body
	| /* empty */
	;

%%

FAWK_API void fbas_error(fawk_ctx_t *ctx, fbas_STYPE tok, const char *msg) { libfawk_error(ctx, msg); }

FAWK_API int fawk_lex_fbas(fbas_STYPE *lval, fawk_ctx_t *ctx);
FAWK_API int fawk_parse_fbas(fawk_ctx_t *ctx) { fawk_parser_loop(fbas_yyctx_t, fbas_STYPE, fawk_lex_fbas, fbas_parse, ctx, fbas_RES_NEXT, fbas_RES_DONE); }
